
from drt_sdk_core.interfaces import INonce


class AccountNonceHolder():
    """
    An abstraction representing an account (user or Smart Contract) on the Network.
    """

    def __init__(self, initial_nonce: INonce = 0):
        self.nonce: INonce = initial_nonce

    def get_nonce_then_increment(self) -> INonce:
        nonce = self.nonce
        self.increment_nonce()
        return nonce

    def increment_nonce(self):
        self.nonce += 1



import pytest

from drt_sdk_core import TokenPayment
from drt_sdk_core.address import Address
from drt_sdk_core.errors import ErrInvalidRelayerV1BuilderArguments
from drt_sdk_core.testutils.wallets import load_wallets
from drt_sdk_core.transaction import Transaction, TransactionComputer
from drt_sdk_core.transaction_builders.relayed_v1_builder import \
    RelayedTransactionV1Builder


class NetworkConfig:
    def __init__(self) -> None:
        self.min_gas_limit = 50_000
        self.gas_per_data_byte = 1_500
        self.gas_price_modifier = 0.01
        self.chain_id = "T"


class TestRelayedV1Builder:
    wallets = load_wallets()
    alice = wallets["alice"]
    bob = wallets["bob"]
    frank = wallets["frank"]
    grace = wallets["grace"]
    carol = wallets["carol"]
    transaction_computer = TransactionComputer()

    def test_without_arguments(self):
        relayed_builder = RelayedTransactionV1Builder()

        with pytest.raises(ErrInvalidRelayerV1BuilderArguments):
            relayed_builder.build()

        inner_transaction = Transaction(
            chain_id="1",
            sender=self.alice.label,
            receiver="moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv",
            gas_limit=10000000,
            nonce=15,
            data=b"getContractConfig"
        )
        relayed_builder.set_inner_transaction(inner_transaction)

        with pytest.raises(ErrInvalidRelayerV1BuilderArguments):
            relayed_builder.build()

        network_config = NetworkConfig()
        relayed_builder.set_network_config(network_config)

        with pytest.raises(ErrInvalidRelayerV1BuilderArguments):
            relayed_builder.build()

    def test_compute_relayed_v1_tx(self):
        network_config = NetworkConfig()

        inner_tx = Transaction(
            chain_id=network_config.chain_id,
            sender=self.bob.label,
            receiver="moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv",
            gas_limit=60000000,
            nonce=198,
            data=b"getContractConfig"
        )
        inner_tx.signature = self.bob.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(inner_tx))

        relayed_builder = RelayedTransactionV1Builder()
        relayed_builder.set_inner_transaction(inner_tx)
        relayed_builder.set_relayer_nonce(2627)
        relayed_builder.set_network_config(network_config)
        relayed_builder.set_relayer_address(Address.new_from_bech32(self.alice.label))

        relayed_tx = relayed_builder.build()
        relayed_tx.signature = self.alice.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(relayed_tx))

        assert relayed_tx.nonce == 2627
        assert relayed_tx.data.decode() == "relayedTx@7b226e6f6e6365223a3139382c2273656e646572223a2267456e574f65576d6d413063306a6b71764d354241707a61644b46574e534f69417643575163776d4750673d222c227265636569766572223a22414141414141414141414141415141414141414141414141414141414141414141414141414141432f2f383d222c2276616c7565223a302c226761735072696365223a313030303030303030302c226761734c696d6974223a36303030303030302c2264617461223a225a3256305132397564484a68593352446232356d6157633d222c227369676e6174757265223a22746264506468326e45757a317973436e414b724c3573386253576e4865304142326434735a4f365253794434315a516f4a69434d346430543762356976587a456475346579595468462f30617a6e41614363354442413d3d222c22636861696e4944223a2256413d3d222c2276657273696f6e223a327d"
        assert relayed_tx.signature.hex() == "71c43095b3289582e8f6fbb1b38fc3a42a2bf091944a4ed64756bf8dd17738dabf9a32da7450e313e7bcec53c6f11d639d6bec79bbd6975841a9517eecd4b207"

    def test_compute_guarded_inner_tx(self):
        network_config = NetworkConfig()

        inner_tx = Transaction(
            chain_id=network_config.chain_id,
            sender=self.bob.label,
            receiver="moa1qqqqqqqqqqqqqpgq54tsxmej537z9leghvp69hfu4f8gg5eu396q2fwu0j",
            gas_limit=60000000,
            nonce=198,
            data=b"getContractConfig",
            guardian=self.grace.label,
            version=2,
            options=2
        )
        inner_tx.signature = self.bob.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(inner_tx))
        inner_tx.guardian_signature = self.grace.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(inner_tx))

        relayed_builder = RelayedTransactionV1Builder()
        relayed_builder.set_inner_transaction(inner_tx)
        relayed_builder.set_relayer_nonce(2627)
        relayed_builder.set_network_config(network_config)
        relayed_builder.set_relayer_address(Address.new_from_bech32(self.alice.label))

        relayed_tx = relayed_builder.build()
        relayed_tx.signature = self.alice.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(relayed_tx))

        assert relayed_tx.nonce == 2627
        assert relayed_tx.data.decode() == "relayedTx@7b226e6f6e6365223a3139382c2273656e646572223a2267456e574f65576d6d413063306a6b71764d354241707a61644b46574e534f69417643575163776d4750673d222c227265636569766572223a22414141414141414141414146414b565841323879704877692f79693741364c64504b704f68464d386958513d222c2276616c7565223a302c226761735072696365223a313030303030303030302c226761734c696d6974223a36303030303030302c2264617461223a225a3256305132397564484a68593352446232356d6157633d222c227369676e6174757265223a22617748304e2b4866704c49346968726374446b56426b33636b575949756656384d7536596e706c7147566239705447584f6f776c375231306a77316f5161545778595573374d41756353784c6e6c616b7378554541773d3d222c22636861696e4944223a2256413d3d222c2276657273696f6e223a322c226f7074696f6e73223a322c22677561726469616e223a22486f714c61306e655733766843716f56696c70715372744c5673774939535337586d7a563868477450684d3d222c22677561726469616e5369676e6174757265223a224252306a414e69706a32796d4466694b426a364e757034454948652f5972383971687a443374453565556670626638503870506159376a44614779514e66735159776f4f4b58414b4873644b5a4557336672427743773d3d227d"
        assert relayed_tx.signature.hex() == "86a1dfafd0a0139f3a0cd24f51f396cd2a573c061c10b13619b2f4ba137c40fbd890559fd65bf941a52cf085244f2af2a139ef2817d15d5c0d1eb7d0dd8a910d"

    def test_guarded_inner_tx_and_guarded_relayed_tx(self):
        network_config = NetworkConfig()

        inner_tx = Transaction(
            chain_id=network_config.chain_id,
            sender=self.bob.label,
            receiver="moa1qqqqqqqqqqqqqpgq54tsxmej537z9leghvp69hfu4f8gg5eu396q2fwu0j",
            gas_limit=60000000,
            nonce=198,
            data=b"addNumber",
            guardian=self.grace.label,
            version=2,
            options=2
        )
        inner_tx.signature = self.bob.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(inner_tx))
        inner_tx.guardian_signature = self.grace.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(inner_tx))

        relayed_builder = RelayedTransactionV1Builder()
        relayed_builder.set_inner_transaction(inner_tx)
        relayed_builder.set_relayer_nonce(2627)
        relayed_builder.set_network_config(network_config)
        relayed_builder.set_relayer_address(Address.new_from_bech32(self.alice.label))
        relayed_builder.set_relayed_transaction_version(2)
        relayed_builder.set_relayed_transaction_options(2)
        relayed_builder.set_relayed_transaction_guardian(Address.new_from_bech32(self.frank.label))

        relayed_tx = relayed_builder.build()
        relayed_tx.signature = self.alice.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(relayed_tx))
        relayed_tx.guardian_signature = self.frank.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(inner_tx))

        assert relayed_tx.nonce == 2627
        assert relayed_tx.data.decode() == "relayedTx@7b226e6f6e6365223a3139382c2273656e646572223a2267456e574f65576d6d413063306a6b71764d354241707a61644b46574e534f69417643575163776d4750673d222c227265636569766572223a22414141414141414141414146414b565841323879704877692f79693741364c64504b704f68464d386958513d222c2276616c7565223a302c226761735072696365223a313030303030303030302c226761734c696d6974223a36303030303030302c2264617461223a225957526b546e5674596d5679222c227369676e6174757265223a226e6a6a7678546353775756396a7a68722b55367041556c3549586d614158762f395648394b4554414a4f493172426b565169536c38442f7962535955634a484b4e6954686243325767796d6b4f564b624a4b6d4442773d3d222c22636861696e4944223a2256413d3d222c2276657273696f6e223a322c226f7074696f6e73223a322c22677561726469616e223a22486f714c61306e655733766843716f56696c70715372744c5673774939535337586d7a563868477450684d3d222c22677561726469616e5369676e6174757265223a22492b4a544753686f696a432f483366557a4c79522f44534b31457135746f79392b796b457554444335746c53765a61577543435a76667a6a704a7872636e5868316e5178453376434a2f7947753836677270583841513d3d227d"
        assert relayed_tx.signature.hex() == "46b0383b43e39948ba78bc63e60dae5d1177b6d5e371578ebcc16f3196a18068b8671c5ff3ca3d42a4f378ec509f72db89a4b55b0b15224f44d53a4fed286604"

    def test_compute_relayedV1_with_usernames(self):
        network_config = NetworkConfig()

        inner_tx = Transaction(
            chain_id=network_config.chain_id,
            sender=self.carol.label,
            receiver=self.alice.label,
            gas_limit=50000,
            sender_username="carol",
            receiver_username="alice",
            nonce=208,
            value=TokenPayment.rewa_from_amount(1).amount_as_integer
        )
        inner_tx.signature = self.carol.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(inner_tx))

        builder = RelayedTransactionV1Builder()
        builder.set_inner_transaction(inner_tx)
        builder.set_relayer_nonce(715)
        builder.set_network_config(network_config)
        builder.set_relayer_address(Address.new_from_bech32(self.frank.label))

        relayed_tx = builder.build()
        relayed_tx.signature = self.frank.secret_key.sign(self.transaction_computer.compute_bytes_for_signing(relayed_tx))

        assert relayed_tx.nonce == 715
        assert relayed_tx.data.decode() == "relayedTx@7b226e6f6e6365223a3230382c2273656e646572223a227371455656633553486b6c45344a717864556e59573068397a536249533141586f3534786f32634969626f3d222c227265636569766572223a2241546c484c76396f686e63616d433877673970645168386b77704742356a6949496f3349484b594e6165453d222c2276616c7565223a313030303030303030303030303030303030302c226761735072696365223a313030303030303030302c226761734c696d6974223a35303030302c2264617461223a22222c227369676e6174757265223a226b702f6f36652b4f666334722f42667545436b446850666330546d52663578622f736665637a487968416f414479763156696a6d6346337a68632f766e7839774c36426e586451384f4c62526e3436537947557743413d3d222c22636861696e4944223a2256413d3d222c2276657273696f6e223a322c22736e64557365724e616d65223a22593246796232773d222c22726376557365724e616d65223a22595778705932553d227d"
        assert relayed_tx.signature.hex() == "aae9b37b80a61e80a86011b4f61fb50a98d23b49bb4899ca962f1c085234a1cb45dfeb509bfcba075c32452d6906b227f99d1c4ffb946df542c644b4b70a9808"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: transaction.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11transaction.proto\x12\x05proto\"\x9a\x02\n\x0bTransaction\x12\r\n\x05Nonce\x18\x01 \x01(\x04\x12\r\n\x05Value\x18\x02 \x01(\x0c\x12\x0f\n\x07RcvAddr\x18\x03 \x01(\x0c\x12\x13\n\x0bRcvUserName\x18\x04 \x01(\x0c\x12\x0f\n\x07SndAddr\x18\x05 \x01(\x0c\x12\x13\n\x0bSndUserName\x18\x06 \x01(\x0c\x12\x10\n\x08GasPrice\x18\x07 \x01(\x04\x12\x10\n\x08GasLimit\x18\x08 \x01(\x04\x12\x0c\n\x04\x44\x61ta\x18\t \x01(\x0c\x12\x0f\n\x07\x43hainID\x18\n \x01(\x0c\x12\x0f\n\x07Version\x18\x0b \x01(\r\x12\x11\n\tSignature\x18\x0c \x01(\x0c\x12\x0f\n\x07Options\x18\r \x01(\r\x12\x11\n\tGuardAddr\x18\x0e \x01(\x0c\x12\x16\n\x0eGuardSignature\x18\x0f \x01(\x0c\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'transaction_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _TRANSACTION._serialized_start = 29  # pyright: ignore[reportUndefinedVariable]
    _TRANSACTION._serialized_end = 311  # pyright: ignore[reportUndefinedVariable]
# @@protoc_insertion_point(module_scope)



from typing import Any


class ErrBadPubkeyLength(Exception):
    def __init__(self, actual: int, expected: int) -> None:
        super().__init__(f"Bad pubkey length: actual = {actual}, expected = {expected}")


class ErrBadAddress(Exception):
    def __init__(self, address: Any) -> None:
        super().__init__(f"Bad address: {address}")


class ErrCannotSerializeArgument(Exception):
    def __init__(self, arg: Any) -> None:
        super().__init__(f"Cannot serialize: {arg}")


class ErrInvalidRelayerV1BuilderArguments(Exception):
    def __init__(self) -> None:
        super().__init__("Invalid arguments for relayed v1 builder")


class ErrInvalidRelayerV2BuilderArguments(Exception):
    def __init__(self) -> None:
        super().__init__("Invalid arguments for relayed v2 builder")


class ErrInvalidGasLimitForInnerTransaction(Exception):
    def __init__(self) -> None:
        super().__init__("Gas limit should be 0 for the inner transaction for relayed v2")


class ErrListsLengthMismatch(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class NotEnoughGasError(Exception):
    def __init__(self, gas_limit: int) -> None:
        super().__init__(f"Not enough gas provided: {gas_limit}")


class BadUsageError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidTokenIdentifierError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidInnerTransactionError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)



from typing import List, Sequence

from drt_sdk_core.interfaces import IAddress
from drt_sdk_core.transaction_parsers.interfaces import (
    IContractResultItem, IContractResults, ITransactionEvent,
    ITransactionEventTopic, ITransactionLogs, ITransactionOnNetwork)


class TransactionOnNetworkWrapper:
    def __init__(self, contract_results: IContractResults, logs: ITransactionLogs):
        self.contract_results: ContractResultsWrapper = ContractResultsWrapper(contract_results.items)
        self.logs: TransactionLogsWrapper = TransactionLogsWrapper(logs.events)

    @classmethod
    def from_transaction(cls, transaction_on_network: ITransactionOnNetwork) -> 'TransactionOnNetworkWrapper':
        return cls(transaction_on_network.contract_results, transaction_on_network.logs)

    def ensure_no_error(self):
        all_events = self.gather_all_events()

        for event in all_events:
            if event.identifier == "signalError":
                data = event.data[1:]
                message = str(event.topics[1])

                raise Exception(f"encountered signalError: {message} ({data})")

    def find_single_event_by_identifier(self, identifier: str) -> 'TransactionEventWrapper':
        all_events = self.gather_all_events()
        filtered_events = [event for event in all_events if event.identifier == identifier]

        if len(filtered_events) == 0:
            raise Exception(f"cannot find event of type: '{identifier}'")
        if len(filtered_events) > 1:
            raise Exception(f"more than one event of type '{identifier}'")

        return filtered_events[0]

    def gather_all_events(self) -> List['TransactionEventWrapper']:
        all_events: List[TransactionEventWrapper] = []
        all_events.extend(self.logs.events)

        for item in self.contract_results.items:
            all_events.extend(item.logs.events)

        return all_events


class ContractResultsWrapper:
    def __init__(self, items: Sequence[IContractResultItem]):
        self.items: List[ContractResultItemWrapper] = [ContractResultItemWrapper(item.logs) for item in items]


class ContractResultItemWrapper:
    def __init__(self, logs: 'ITransactionLogs'):
        self.logs: TransactionLogsWrapper = TransactionLogsWrapper(logs.events)


class TransactionLogsWrapper:
    def __init__(self, events: Sequence[ITransactionEvent]):
        self.events: List[TransactionEventWrapper] = [TransactionEventWrapper(event.address, event.identifier, event.topics, event.data) for event in events]


class TransactionEventWrapper:
    def __init__(self, address: IAddress, identifier: str, topics: Sequence[ITransactionEventTopic], data: str):
        self.address: IAddress = address
        self.identifier: str = identifier
        self.topics: List[TransactionEventTopicWrapper] = [TransactionEventTopicWrapper(topic.raw) for topic in topics]
        self.data: str = data


class TransactionEventTopicWrapper:
    def __init__(self, raw: bytes):
        self.raw = raw

    def __str__(self) -> str:
        return self.raw.decode("utf-8")



from typing import List

from drt_sdk_core.codec import (decode_unsigned_number,
                                       encode_unsigned_number)
from drt_sdk_core.constants import TOKEN_RANDOM_SEQUENCE_LENGTH
from drt_sdk_core.errors import (BadUsageError,
                                        InvalidTokenIdentifierError)
from drt_sdk_core.interfaces import IToken, ITokenIdentifierParts


class Token:
    def __init__(self, identifier: str, nonce: int = 0) -> None:
        self.identifier = identifier
        self.nonce = nonce


class TokenTransfer:
    def __init__(self, token: IToken, amount: int) -> None:
        """`amount` should always be in atomic units: 1.000000 "USDC-c76f1f" = "1000000"""
        self.token = token
        self.amount = amount


class TokenIdentifierParts:
    def __init__(self, ticker: str, random_sequence: str, nonce: int) -> None:
        self.ticker = ticker
        self.random_sequence = random_sequence
        self.nonce = nonce


class TokenComputer:
    def __init__(self) -> None:
        pass

    def is_fungible(self, token: IToken) -> bool:
        return token.nonce == 0

    def extract_nonce_from_extended_identifier(self, identifier: str) -> int:
        parts = identifier.split("-")

        self._check_if_extended_identifier_was_provided(parts)
        self._check_length_of_random_sequence(parts[1])

        # in case the identifier of a fungible token is provided
        if len(parts) == 2:
            return 0

        hex_nonce = bytes.fromhex(parts[2])
        return decode_unsigned_number(hex_nonce)

    def extract_identifier_from_extended_identifier(self, identifier: str) -> str:
        parts = identifier.split("-")

        self._check_if_extended_identifier_was_provided(parts)
        self._ensure_token_ticker_validity(parts[0])
        self._check_length_of_random_sequence(parts[1])

        return parts[0] + "-" + parts[1]

    def extract_ticker_from_identifier(self, identifier: str) -> str:
        parts = identifier.split("-")

        self._check_length_of_random_sequence(parts[1])
        self._ensure_token_ticker_validity(parts[0])
        return parts[0]

    def parse_extended_identifier_parts(self, identifier: str) -> TokenIdentifierParts:
        parts = identifier.split("-")

        self._check_if_extended_identifier_was_provided(parts)
        self._check_length_of_random_sequence(parts[1])
        self._ensure_token_ticker_validity(parts[0])

        nonce = decode_unsigned_number(bytes.fromhex(parts[2])) if len(parts) == 3 else 0
        return TokenIdentifierParts(parts[0], parts[1], nonce)

    def compute_extended_identifier_from_identifier_and_nonce(self, identifier: str, nonce: int) -> str:
        identifier_parts = identifier.split("-")

        self._check_length_of_random_sequence(identifier_parts[1])
        self._ensure_token_ticker_validity(identifier_parts[0])

        if nonce < 0:
            raise BadUsageError("The token nonce can not be less than 0")

        if nonce == 0:
            return identifier

        nonce_hex = encode_unsigned_number(nonce).hex()
        return identifier + "-" + nonce_hex

    def compute_extended_identifier_from_parts(self, parts: ITokenIdentifierParts) -> str:
        identifier = parts.ticker + "-" + parts.random_sequence
        return self.compute_extended_identifier_from_identifier_and_nonce(identifier, parts.nonce)

    def _check_if_extended_identifier_was_provided(self, token_parts: List[str]) -> None:
        # this is for the identifiers of fungible tokens
        MIN_EXTENDED_IDENTIFIER_LENGTH_IF_SPLIT = 2
        # this is for the identifiers of nft, sft and meta-dcdt
        MAX_EXTENDED_IDENTIFIER_LENGTH_IF_SPLIT = 3

        if len(token_parts) < MIN_EXTENDED_IDENTIFIER_LENGTH_IF_SPLIT or len(token_parts) > MAX_EXTENDED_IDENTIFIER_LENGTH_IF_SPLIT:
            raise InvalidTokenIdentifierError("Invalid extended token identifier provided")

    def _ensure_token_ticker_validity(self, ticker: str) -> None:
        MIN_TICKER_LENGTH = 3
        MAX_TICKER_LENGTH = 10

        if len(ticker) < MIN_TICKER_LENGTH or len(ticker) > MAX_TICKER_LENGTH:
            raise InvalidTokenIdentifierError(f"The token ticker should be between {MIN_TICKER_LENGTH} and {MAX_TICKER_LENGTH} characters")

        if not ticker.isalnum():
            raise InvalidTokenIdentifierError("The token ticker should only contain alphanumeric characters")

        if not ticker.isupper():
            raise InvalidTokenIdentifierError("The token ticker should be upper case")

    def _check_length_of_random_sequence(self, random_sequence: str) -> None:
        if len(random_sequence) != TOKEN_RANDOM_SEQUENCE_LENGTH:
            raise InvalidTokenIdentifierError("The identifier is not valid. The random sequence does not have the right length")



from typing import List

from drt_sdk_core.address import Address
from drt_sdk_core.code_metadata import CodeMetadata
from drt_sdk_core.interfaces import ITokenPayment
from drt_sdk_core.token_payment import TokenPayment
from drt_sdk_core.transaction_builders.contract_builders import (
    ContractCallBuilder, ContractDeploymentBuilder, ContractUpgradeBuilder)
from drt_sdk_core.transaction_builders.default_configuration import \
    DefaultTransactionBuildersConfiguration

dummyConfig = DefaultTransactionBuildersConfiguration(chain_id="D")


def test_contract_deployment_builder():
    owner = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
    metadata = CodeMetadata(upgradeable=True, readable=True, payable=True, payable_by_contract=True)

    builder = ContractDeploymentBuilder(
        dummyConfig,
        owner=owner,
        deploy_arguments=[42, "test"],
        code_metadata=metadata,
        code=bytes([0xAA, 0xBB, 0xCC, 0xDD]),
        gas_limit=10000000
    )

    payload = builder.build_payload()
    tx = builder.build()

    assert payload.data == b"aabbccdd@0500@0506@2a@74657374"
    assert tx.chain_id == "D"
    assert tx.sender == owner.to_bech32()
    assert tx.receiver == "moa1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqhsx6tv"
    assert tx.gas_limit == 10000000
    assert tx.gas_price == 1000000000
    assert tx.data.decode() == str(payload)


def test_contract_upgrade_builder():
    contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgquzmh78klkqwt0p4rjys0qtp3la07gz4d396q7vfu0t")
    owner = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
    metadata = CodeMetadata(upgradeable=True, readable=True, payable=True, payable_by_contract=True)

    builder = ContractUpgradeBuilder(
        dummyConfig,
        contract=contract,
        owner=owner,
        upgrade_arguments=[42, "test"],
        code_metadata=metadata,
        code=bytes([0xAA, 0xBB, 0xCC, 0xDD]),
        gas_limit=10000000
    )

    payload = builder.build_payload()
    tx = builder.build()

    assert payload.data == b"upgradeContract@aabbccdd@0506@2a@74657374"
    assert tx.chain_id == "D"
    assert tx.sender == owner.to_bech32()
    assert tx.receiver == contract.to_bech32()
    assert tx.gas_limit == 10000000
    assert tx.gas_price == 1000000000
    assert tx.data.decode() == str(payload)


def test_contract_call_builder():
    contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgquzmh78klkqwt0p4rjys0qtp3la07gz4d396q7vfu0t")
    caller = Address.new_from_bech32("moa1k2s324ww2g0yj38qn2ch2jwctdy8mnfxep94q9arncc6xecg3xaqhr5l9h")

    builder = ContractCallBuilder(
        dummyConfig,
        contract=contract,
        function_name="foo",
        caller=caller,
        call_arguments=[42, "test", Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")],
        gas_limit=10000000
    )

    payload = builder.build_payload()
    tx = builder.build()

    assert payload.data == b"foo@2a@74657374@0139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1"
    assert tx.sender == caller.to_bech32()
    assert tx.receiver == contract.to_bech32()
    assert tx.gas_limit == 10000000
    assert tx.data.decode() == str(payload)


def test_contract_call_builder_with_dcdt_transfer():
    contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgquzmh78klkqwt0p4rjys0qtp3la07gz4d396q7vfu0t")
    caller = Address.new_from_bech32("moa1k2s324ww2g0yj38qn2ch2jwctdy8mnfxep94q9arncc6xecg3xaqhr5l9h")

    transfers: List[ITokenPayment] = [
        TokenPayment.fungible_from_amount("COUNTER-8b028f", "100.00", 2)
    ]

    builder = ContractCallBuilder(
        dummyConfig,
        contract=contract,
        function_name="hello",
        caller=caller,
        call_arguments=[42, "test", Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")],
        gas_limit=10000000,
        dcdt_transfers=transfers
    )

    payload = builder.build_payload()
    tx = builder.build()

    assert payload.data == b"DCDTTransfer@434f554e5445522d386230323866@2710@68656c6c6f@2a@74657374@0139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1"
    assert tx.sender == caller.to_bech32()
    assert tx.receiver == contract.to_bech32()
    assert tx.data.decode() == str(payload)


def test_contract_call_builder_with_dcdt_nft_transfer():
    contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgquzmh78klkqwt0p4rjys0qtp3la07gz4d396q7vfu0t")
    caller = Address.new_from_bech32("moa1k2s324ww2g0yj38qn2ch2jwctdy8mnfxep94q9arncc6xecg3xaqhr5l9h")

    transfers: List[ITokenPayment] = [
        TokenPayment.non_fungible("TEST-38f249", 1)
    ]

    builder = ContractCallBuilder(
        dummyConfig,
        contract=contract,
        function_name="hello",
        caller=caller,
        call_arguments=[42, "test", Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")],
        gas_limit=10000000,
        dcdt_transfers=transfers
    )

    payload = builder.build_payload()
    tx = builder.build()

    assert payload.data == b"DCDTNFTTransfer@544553542d333866323439@01@01@00000000000000000500e0b77f1edfb01cb786a39120f02c31ff5fe40aad8974@68656c6c6f@2a@74657374@0139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1"
    assert tx.sender == caller.to_bech32()
    assert tx.receiver == caller.to_bech32()
    assert tx.data.decode() == str(payload)


def test_contract_call_builder_with_multi_dcdt_nft_transfer():
    contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgquzmh78klkqwt0p4rjys0qtp3la07gz4d396q7vfu0t")
    caller = Address.new_from_bech32("moa1k2s324ww2g0yj38qn2ch2jwctdy8mnfxep94q9arncc6xecg3xaqhr5l9h")

    transfers = [
        TokenPayment.non_fungible("TEST-38f249", 1),
        TokenPayment.fungible_from_amount("BAR-c80d29", "10.00", 18)
    ]

    builder = ContractCallBuilder(
        dummyConfig,
        contract=contract,
        function_name="hello",
        caller=caller,
        call_arguments=[42, "test", Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")],
        gas_limit=10000000,
        dcdt_transfers=transfers
    )

    payload = builder.build_payload()
    tx = builder.build()

    assert payload.data == b"MultiDCDTNFTTransfer@00000000000000000500e0b77f1edfb01cb786a39120f02c31ff5fe40aad8974@02@544553542d333866323439@01@01@4241522d633830643239@@8ac7230489e80000@68656c6c6f@2a@74657374@0139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1"
    assert tx.sender == caller.to_bech32()
    assert tx.receiver == caller.to_bech32()
    assert tx.data.decode() == str(payload)



from typing import Any, List, Optional, Protocol, Sequence

from drt_sdk_core.constants import VM_TYPE_WASM_VM
from drt_sdk_core.interfaces import (IAddress, ICodeMetadata, IGasLimit,
                                            IGasPrice, INonce, ITokenPayment,
                                            ITransactionValue)
from drt_sdk_core.serializer import arg_to_string, args_to_strings
from drt_sdk_core.transaction_builders.transaction_builder import (
    ITransactionBuilderConfiguration, TransactionBuilder)


class IContractDeploymentConfiguration(ITransactionBuilderConfiguration, Protocol):
    deployment_address: IAddress


class ContractDeploymentBuilder(TransactionBuilder):
    def __init__(self,
                 config: IContractDeploymentConfiguration,
                 code: bytes,
                 code_metadata: ICodeMetadata,
                 deploy_arguments: Sequence[Any],
                 owner: IAddress,
                 nonce: Optional[INonce] = None,
                 value: Optional[ITransactionValue] = None,
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None
                 ) -> None:
        super().__init__(config, nonce, value, gas_limit, gas_price)
        self.sender = owner
        self.receiver = config.deployment_address
        self.code = code
        self.code_metadata = code_metadata
        self.deploy_arguments = deploy_arguments

    def _build_payload_parts(self) -> List[str]:
        return [
            self.code.hex(),
            arg_to_string(VM_TYPE_WASM_VM),
            arg_to_string(self.code_metadata)
        ] + args_to_strings(self.deploy_arguments)


class ContractUpgradeBuilder(TransactionBuilder):
    def __init__(self,
                 config: ITransactionBuilderConfiguration,
                 contract: IAddress,
                 code: bytes,
                 code_metadata: ICodeMetadata,
                 upgrade_arguments: Sequence[Any],
                 owner: IAddress,
                 nonce: Optional[INonce] = None,
                 value: Optional[ITransactionValue] = None,
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None
                 ) -> None:
        super().__init__(config, nonce, value, gas_limit, gas_price)
        self.sender = owner
        self.receiver = contract
        self.code = code
        self.code_metadata = code_metadata
        self.upgrade_arguments = upgrade_arguments
        self.owner = owner

    def _build_payload_parts(self) -> List[str]:
        return [
            "upgradeContract",
            self.code.hex(),
            arg_to_string(self.code_metadata)
        ] + args_to_strings(self.upgrade_arguments)


class ContractCallBuilder(TransactionBuilder):
    def __init__(self,
                 config: ITransactionBuilderConfiguration,
                 contract: IAddress,
                 function_name: str,
                 call_arguments: Sequence[Any],
                 caller: IAddress,
                 nonce: Optional[INonce] = None,
                 value: Optional[ITransactionValue] = None,
                 dcdt_transfers: Sequence[ITokenPayment] = [],
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None
                 ) -> None:
        super().__init__(config, nonce, value, gas_limit, gas_price)
        self.contract = contract
        self.function_name = function_name
        self.call_arguments = call_arguments
        self.caller = caller
        self.dcdt_transfers = dcdt_transfers

    def _get_sender(self) -> IAddress:
        return self.caller

    def _get_receiver(self) -> IAddress:
        receiver_is_same_as_sender = self._has_multiple_transfers() or self._has_single_nft_transfer()
        return self.caller if receiver_is_same_as_sender else self.contract

    def _build_payload_parts(self) -> List[str]:
        parts: List[str] = []

        if self._has_single_dcdt_transfer():
            transfer = self.dcdt_transfers[0]
            parts = [
                "DCDTTransfer",
                arg_to_string(transfer.token_identifier),
                arg_to_string(transfer.amount_as_integer),
                arg_to_string(self.function_name),
                *args_to_strings(self.call_arguments)
            ]
        elif self._has_single_nft_transfer():
            transfer = self.dcdt_transfers[0]
            parts = [
                "DCDTNFTTransfer",
                arg_to_string(transfer.token_identifier),
                arg_to_string(transfer.token_nonce),
                arg_to_string(transfer.amount_as_integer),
                self.contract.to_hex(),
                arg_to_string(self.function_name),
                *args_to_strings(self.call_arguments)
            ]
        elif self._has_multiple_transfers():
            parts = [
                "MultiDCDTNFTTransfer",
                self.contract.to_hex(),
                arg_to_string(len(self.dcdt_transfers))
            ]

            for transfer in self.dcdt_transfers:
                parts.extend([
                    arg_to_string(transfer.token_identifier),
                    arg_to_string(transfer.token_nonce),
                    arg_to_string(transfer.amount_as_integer)
                ])

            parts.extend([
                arg_to_string(self.function_name),
                *args_to_strings(self.call_arguments)
            ])
        else:
            parts = [self.function_name] + args_to_strings(self.call_arguments)

        return parts

    def _has_single_dcdt_transfer(self) -> bool:
        return len(self.dcdt_transfers) == 1 and self.dcdt_transfers[0].is_fungible()

    def _has_single_nft_transfer(self) -> bool:
        return len(self.dcdt_transfers) == 1 and not self.dcdt_transfers[0].is_fungible()

    def _has_multiple_transfers(self) -> bool:
        return len(self.dcdt_transfers) > 1



TRANSACTION_MIN_GAS_PRICE = 1000000000
TRANSACTION_VERSION_DEFAULT = 2
TRANSACTION_OPTIONS_DEFAULT = 0
ARGS_SEPARATOR = "@"
METACHAIN_ID = 4294967295
VM_TYPE_WASM_VM = bytes([0x05, 0x00])

# 64 bytes = 512 bits
INTEGER_MAX_NUM_BYTES = 64

REWA_TOKEN_IDENTIFIER = "REWA"
REWA_NUM_DECIMALS = 18

DELEGATION_MANAGER_SC_ADDRESS = "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqylllsjrx4c2"
DEFAULT_HRP = "moa"
CONTRACT_DEPLOY_ADDRESS = "moa1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqhsx6tv"
TRANSACTION_OPTIONS_TX_GUARDED = 0b0010

DIGEST_SIZE = 32

TOKEN_RANDOM_SEQUENCE_LENGTH = 6



import logging
from typing import List, Optional, Protocol

from drt_sdk_core.constants import (ARGS_SEPARATOR,
                                           TRANSACTION_OPTIONS_DEFAULT,
                                           TRANSACTION_VERSION_DEFAULT)
from drt_sdk_core.interfaces import (IAddress, IChainID, IGasLimit,
                                            IGasPrice, INonce,
                                            ITransactionOptions,
                                            ITransactionPayload,
                                            ITransactionValue,
                                            ITransactionVersion)
from drt_sdk_core.transaction import Transaction
from drt_sdk_core.transaction_payload import TransactionPayload

logger = logging.getLogger("transaction_builder")


class ITransactionBuilderConfiguration(Protocol):
    chain_id: IChainID
    min_gas_price: IGasPrice
    min_gas_limit: IGasLimit
    gas_limit_per_byte: IGasLimit


class TransactionBuilder:
    """
    Not intended to be used directly; should be derived and specialized.
    """

    def __init__(self,
                 config: ITransactionBuilderConfiguration,
                 nonce: Optional[INonce] = None,
                 value: Optional[ITransactionValue] = None,
                 gas_limit: Optional[IGasLimit] = None,
                 gas_price: Optional[IGasPrice] = None) -> None:
        logger.warning("The transaction builders are deprecated and will be removed in a later release. Use the transaction factories instead.")

        self.chain_id = config.chain_id
        self.min_gas_price = config.min_gas_price
        self.min_gas_limit = config.min_gas_limit
        self.gas_limit_per_byte = config.gas_limit_per_byte

        self.nonce = nonce
        self.value = value
        self.gas_limit = gas_limit
        self.gas_price = gas_price or config.min_gas_price
        self.sender: Optional[IAddress] = None
        self.receiver: Optional[IAddress] = None

    def build(self) -> Transaction:
        chain_id = self.chain_id
        sender = self._get_sender()
        receiver = self._get_receiver()
        data = self.build_payload()
        gas_limit = self._get_gas_limit(data)
        gas_price = self.gas_price
        nonce = self.nonce or 0
        value = self._get_value()
        version = self._get_transaction_version()
        options = self._get_transaction_options()

        return Transaction(
            chain_id=chain_id,
            sender=sender.to_bech32(),
            receiver=receiver.to_bech32(),
            gas_limit=gas_limit,
            gas_price=gas_price,
            nonce=nonce,
            value=int(str(value)),
            data=data.data,
            version=version,
            options=options
        )

    def build_payload(self) -> TransactionPayload:
        parts = self._build_payload_parts()
        data = ARGS_SEPARATOR.join(parts)
        payload = TransactionPayload.from_str(data)
        return payload

    def _build_payload_parts(self) -> List[str]:
        raise NotImplementedError()

    def _get_sender(self) -> IAddress:
        assert self.sender is not None, "sender isn't set"
        return self.sender

    def _get_receiver(self) -> IAddress:
        assert self.receiver is not None, "receiver isn't set"
        return self.receiver

    def _get_value(self) -> ITransactionValue:
        return self.value or 0

    def _get_gas_limit(self, payload: ITransactionPayload) -> IGasLimit:
        if self.gas_limit:
            return self.gas_limit

        gas_limit = self._compute_data_movement_gas(payload) + self._estimate_execution_gas()
        return gas_limit

    def _compute_data_movement_gas(self, payload: ITransactionPayload) -> IGasLimit:
        return self.min_gas_limit + self.gas_limit_per_byte * payload.length()

    def _estimate_execution_gas(self) -> IGasLimit:
        raise NotImplementedError()

    def _get_transaction_version(self) -> ITransactionVersion:
        return TRANSACTION_VERSION_DEFAULT

    def _get_transaction_options(self) -> ITransactionOptions:
        return TRANSACTION_OPTIONS_DEFAULT



from pathlib import Path

from drt_sdk_core.address import Address
from drt_sdk_core.constants import CONTRACT_DEPLOY_ADDRESS
from drt_sdk_core.tokens import Token, TokenComputer, TokenTransfer
from drt_sdk_core.transaction_factories.smart_contract_transactions_factory import \
    SmartContractTransactionsFactory
from drt_sdk_core.transaction_factories.transactions_factory_config import \
    TransactionsFactoryConfig


class TestSmartContractTransactionsFactory:
    config = TransactionsFactoryConfig("D")
    factory = SmartContractTransactionsFactory(config, TokenComputer())

    def test_create_transaction_for_deploy(self):
        sender = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        contract = Path(__file__).parent.parent / "testutils" / "testdata" / "adder.wasm"
        gas_limit = 6000000
        args = [0]

        transaction = self.factory.create_transaction_for_deploy(
            sender=sender,
            bytecode=contract,
            gas_limit=gas_limit,
            arguments=args
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == CONTRACT_DEPLOY_ADDRESS
        assert transaction.data
        assert transaction.gas_limit == gas_limit
        assert transaction.value == 0

    def test_create_transaction_for_execute_no_transfer(self):
        sender = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9")
        function = "add"
        gas_limit = 6000000
        args = [7]

        intent = self.factory.create_transaction_for_execute(
            sender=sender,
            contract=contract,
            function=function,
            gas_limit=gas_limit,
            arguments=args
        )

        assert intent.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert intent.receiver == "moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9"
        assert intent.gas_limit == gas_limit
        assert intent.data
        assert intent.data.decode() == "add@07"
        assert intent.value == 0

    def test_create_transaction_for_execute_and_tranfer_native_token(self):
        sender = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9")
        function = "add"
        gas_limit = 6000000
        args = [7]
        rewa_amount = 1000000000000000000

        transaction = self.factory.create_transaction_for_execute(
            sender=sender,
            contract=contract,
            function=function,
            gas_limit=gas_limit,
            arguments=args,
            native_transfer_amount=rewa_amount
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9"
        assert transaction.gas_limit == gas_limit
        assert transaction.data
        assert transaction.data.decode() == "add@07"
        assert transaction.value == 1000000000000000000

    def test_create_transaction_for_execute_and_send_single_dcdt(self):
        sender = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9")
        function = "dummy"
        gas_limit = 6000000
        args = [7]
        token = Token("FOO-6ce17b", 0)
        transfer = TokenTransfer(token, 10)

        transaction = self.factory.create_transaction_for_execute(
            sender=sender,
            contract=contract,
            function=function,
            gas_limit=gas_limit,
            arguments=args,
            token_transfers=[transfer]
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9"
        assert transaction.gas_limit == gas_limit
        assert transaction.data
        assert transaction.data.decode() == "DCDTTransfer@464f4f2d366365313762@0a@64756d6d79@07"
        assert transaction.value == 0

    def test_create_transaction_for_execute_and_send_multiple_dcdts(self):
        sender = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgqak8zt22wl2ph4tswtyc39namqx6ysa2sd8ssc7aswp")
        function = "dummy"
        gas_limit = 6000000
        args = [7]

        foo_token = Token("FOO-6ce17b", 0)
        foo_transfer = TokenTransfer(foo_token, 10)

        bar_token = Token("BAR-5bc08f", 0)
        bar_transfer = TokenTransfer(bar_token, 3140)

        transaction = self.factory.create_transaction_for_execute(
            sender=sender,
            contract=contract,
            function=function,
            gas_limit=gas_limit,
            arguments=args,
            token_transfers=[foo_transfer, bar_transfer]
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.gas_limit == gas_limit
        assert transaction.data
        assert transaction.data.decode() == "MultiDCDTNFTTransfer@00000000000000000500ed8e25a94efa837aae0e593112cfbb01b448755069e1@02@464f4f2d366365313762@@0a@4241522d356263303866@@0c44@64756d6d79@07"
        assert transaction.value == 0

    def test_create_transaction_for_execute_and_send_single_nft(self):
        sender = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9")
        function = "dummy"
        gas_limit = 6000000
        args = [7]
        token = Token("NFT-123456", 1)
        transfer = TokenTransfer(token, 1)

        transaction = self.factory.create_transaction_for_execute(
            sender=sender,
            contract=contract,
            function=function,
            gas_limit=gas_limit,
            arguments=args,
            token_transfers=[transfer]
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.gas_limit == gas_limit
        assert transaction.data
        assert transaction.data.decode() == "DCDTNFTTransfer@4e46542d313233343536@01@01@00000000000000000500b9353fe8407f87310c87e12fa1ac807f0485da39d152@64756d6d79@07"
        assert transaction.value == 0

    def test_create_transaction_for_execute_and_send_multiple_nfts(self):
        sender = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9")
        function = "dummy"
        gas_limit = 6000000
        args = [7]

        first_token = Token("NFT-123456", 1)
        first_transfer = TokenTransfer(first_token, 1)
        second_token = Token("NFT-123456", 42)
        second_transfer = TokenTransfer(second_token, 1)

        transaction = self.factory.create_transaction_for_execute(
            sender=sender,
            contract=contract,
            function=function,
            gas_limit=gas_limit,
            arguments=args,
            token_transfers=[first_transfer, second_transfer]
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.gas_limit == gas_limit
        assert transaction.data
        assert transaction.data.decode() == "MultiDCDTNFTTransfer@00000000000000000500b9353fe8407f87310c87e12fa1ac807f0485da39d152@02@4e46542d313233343536@01@01@4e46542d313233343536@2a@01@64756d6d79@07"
        assert transaction.value == 0

    def test_create_transaction_for_upgrade(self):
        sender = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        contract_address = Address.new_from_bech32("moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9")
        contract = Path(__file__).parent.parent / "testutils" / "testdata" / "adder.wasm"
        gas_limit = 6000000
        args = [0]

        transaction = self.factory.create_transaction_for_upgrade(
            sender=sender,
            contract=contract_address,
            bytecode=contract,
            gas_limit=gas_limit,
            arguments=args
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqpgqhy6nl6zq07rnzry8uyh6rtyq0uzgtk3e69fq9ny2r9"
        assert transaction.data
        assert transaction.data.decode().startswith("upgradeContract@")
        assert transaction.gas_limit == gas_limit
        assert transaction.value == 0



from typing import Any, List, Protocol, Sequence, runtime_checkable

from drt_sdk_core.codec import encode_unsigned_number, encode_signed_number
from drt_sdk_core.constants import ARGS_SEPARATOR
from drt_sdk_core.errors import ErrCannotSerializeArgument


@runtime_checkable
class IArgument(Protocol):
    def serialize(self) -> bytes: ...


def args_to_string(args: Sequence[Any]) -> str:
    strings = args_to_strings(args)
    return ARGS_SEPARATOR.join(strings)


def args_to_strings(args: Sequence[Any]) -> List[str]:
    buffers = args_to_buffers(args)
    return [buffer.hex() for buffer in buffers]


def args_to_buffers(args: Sequence[Any]) -> List[bytes]:
    return [arg_to_buffer(arg) for arg in args]


def arg_to_string(arg: Any) -> str:
    buffer = arg_to_buffer(arg)
    return buffer.hex()


def arg_to_buffer(arg: Any) -> bytes:
    if isinstance(arg, str):
        return arg.encode("utf-8")
    if isinstance(arg, int):
        if arg < 0:
            return encode_signed_number(arg)
        return encode_unsigned_number(arg)
    if isinstance(arg, bytes):
        return arg
    if isinstance(arg, bytearray):
        return bytes(arg)
    if isinstance(arg, IArgument):
        return arg.serialize()
    raise ErrCannotSerializeArgument(arg)





from typing import Any, Optional, Sequence

from drt_sdk_core.contract_query import ContractQuery
from drt_sdk_core.interfaces import IAddress
from drt_sdk_core.serializer import args_to_strings


class ContractQueryBuilder():
    def __init__(self,
                 contract: IAddress,
                 function: str,
                 call_arguments: Sequence[Any] = [],
                 caller: Optional[IAddress] = None,
                 value: Optional[int] = None,
                 ) -> None:
        self.contract = contract
        self.function_name = function
        self.call_arguments = call_arguments
        self.caller = caller
        self.value = value

    def build(self) -> ContractQuery:
        query = ContractQuery(
            contract=self.contract,
            function=self.function_name,
            encoded_arguments=args_to_strings(self.call_arguments),
            caller=self.caller,
            value=self.value
        )

        return query



from pathlib import Path
from typing import Dict

from drt_sdk_wallet import UserPEM

alice_pem = Path(__file__).parent / "testwallets" / "alice.pem"
bob_pem = Path(__file__).parent / "testwallets" / "bob.pem"
frank_pem = Path(__file__).parent / "testwallets" / "frank.pem"
grace_pem = Path(__file__).parent / "testwallets" / "grace.pem"
carol_pem = Path(__file__).parent / "testwallets" / "carol.pem"


def load_wallets() -> Dict[str, UserPEM]:
    wallets: Dict[str, UserPEM] = {}

    alice = UserPEM.from_file(Path(alice_pem))
    wallets["alice"] = alice

    bob = UserPEM.from_file(Path(bob_pem))
    wallets["bob"] = bob

    carol = UserPEM.from_file(Path(carol_pem))
    wallets["carol"] = carol

    frank = UserPEM.from_file(Path(frank_pem))
    wallets["frank"] = frank

    grace = UserPEM.from_file(Path(grace_pem))
    wallets["grace"] = grace

    return wallets



from drt_sdk_core import Address
from drt_sdk_core.transaction_factories.token_management_transactions_factory import (
    RegisterAndSetAllRolesTokenType, TokenManagementTransactionsFactory)
from drt_sdk_core.transaction_factories.transactions_factory_config import \
    TransactionsFactoryConfig

frank = Address.new_from_bech32("moa1kdl46yctawygtwg2k462307dmz2v55c605737dp3zkxh04sct7asdueccu")
grace = Address.new_from_bech32("moa1r69gk66fmedhhcg24g2c5kn2f2a5k4kvpr6jfw67dn2lyydd8cfsruuk3f")
alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
factory = TokenManagementTransactionsFactory(TransactionsFactoryConfig("T"))


def test_create_transaction_for_registering_and_setting_roles():
    transaction = factory.create_transaction_for_registering_and_setting_roles(
        sender=frank,
        token_name="TEST",
        token_ticker="TEST",
        token_type=RegisterAndSetAllRolesTokenType.FNG,
        num_decimals=2
    )

    assert transaction.data
    assert transaction.data.decode() == "registerAndSetAllRoles@54455354@54455354@464e47@02"
    assert transaction.sender == frank.to_bech32()
    assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv"
    assert transaction.value == 50000000000000000


def test_create_transaction_for_issuing_fungible():
    transaction = factory.create_transaction_for_issuing_fungible(
        sender=frank,
        token_name="FRANK",
        token_ticker="FRANK",
        initial_supply=100,
        num_decimals=0,
        can_freeze=True,
        can_wipe=True,
        can_pause=True,
        can_change_owner=True,
        can_upgrade=False,
        can_add_special_roles=False
    )

    assert transaction.data
    assert transaction.data.decode() == "issue@4652414e4b@4652414e4b@64@@63616e467265657a65@74727565@63616e57697065@74727565@63616e5061757365@74727565@63616e4368616e67654f776e6572@74727565@63616e55706772616465@66616c7365@63616e4164645370656369616c526f6c6573@66616c7365"
    assert transaction.sender == frank.to_bech32()
    assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv"
    assert transaction.value == 50000000000000000


def test_create_transaction_for_issuing_semi_fungible():
    transaction = factory.create_transaction_for_issuing_semi_fungible(
        sender=frank,
        token_name="FRANK",
        token_ticker="FRANK",
        can_freeze=True,
        can_wipe=True,
        can_pause=True,
        can_transfer_nft_create_role=True,
        can_change_owner=True,
        can_upgrade=False,
        can_add_special_roles=False
    )

    assert transaction.data
    assert transaction.data.decode() == "issueSemiFungible@4652414e4b@4652414e4b@63616e467265657a65@74727565@63616e57697065@74727565@63616e5061757365@74727565@63616e5472616e736665724e4654437265617465526f6c65@74727565@63616e4368616e67654f776e6572@74727565@63616e55706772616465@66616c7365@63616e4164645370656369616c526f6c6573@66616c7365"
    assert transaction.sender == frank.to_bech32()
    assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv"
    assert transaction.value == 50000000000000000


def test_create_transaction_for_issuing_non_fungible():
    transaction = factory.create_transaction_for_issuing_non_fungible(
        sender=frank,
        token_name="FRANK",
        token_ticker="FRANK",
        can_freeze=True,
        can_wipe=True,
        can_pause=True,
        can_transfer_nft_create_role=True,
        can_change_owner=True,
        can_upgrade=False,
        can_add_special_roles=False
    )

    assert transaction.data
    assert transaction.data.decode() == "issueNonFungible@4652414e4b@4652414e4b@63616e467265657a65@74727565@63616e57697065@74727565@63616e5061757365@74727565@63616e5472616e736665724e4654437265617465526f6c65@74727565@63616e4368616e67654f776e6572@74727565@63616e55706772616465@66616c7365@63616e4164645370656369616c526f6c6573@66616c7365"
    assert transaction.sender == frank.to_bech32()
    assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv"
    assert transaction.value == 50000000000000000


def test_create_transaction_for_registering_meta_dcdt():
    transaction = factory.create_transaction_for_registering_meta_dcdt(
        sender=frank,
        token_name="FRANK",
        token_ticker="FRANK",
        num_decimals=10,
        can_freeze=True,
        can_wipe=True,
        can_pause=True,
        can_transfer_nft_create_role=True,
        can_change_owner=True,
        can_upgrade=False,
        can_add_special_roles=False
    )

    assert transaction.data
    assert transaction.data.decode() == "registerMetaDCDT@4652414e4b@4652414e4b@0a@63616e467265657a65@74727565@63616e57697065@74727565@63616e5061757365@74727565@63616e5472616e736665724e4654437265617465526f6c65@74727565@63616e4368616e67654f776e6572@74727565@63616e55706772616465@66616c7365@63616e4164645370656369616c526f6c6573@66616c7365"
    assert transaction.sender == frank.to_bech32()
    assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv"
    assert transaction.value == 50000000000000000


def test_create_transaction_for_setting_special_role_on_non_fungible_token():
    transaction = factory.create_transaction_for_setting_special_role_on_non_fungible_token(
        sender=frank,
        user=grace,
        token_identifier="FRANK-11ce3e",
        add_role_nft_create=True,
        add_role_nft_burn=False,
        add_role_nft_update_attributes=True,
        add_role_nft_add_uri=True,
        add_role_dcdt_transfer_role=False
    )

    assert transaction.data
    assert transaction.data.decode() == "setSpecialRole@4652414e4b2d313163653365@1e8a8b6b49de5b7be10aaa158a5a6a4abb4b56cc08f524bb5e6cd5f211ad3e13@44434454526f6c654e4654437265617465@44434454526f6c654e465455706461746541747472696275746573@44434454526f6c654e4654416464555249"
    assert transaction.sender == frank.to_bech32()
    assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv"
    assert transaction.value == 0


def test_create_transaction_for_creating_nft():
    transaction = factory.create_transaction_for_creating_nft(
        sender=grace,
        token_identifier="FRANK-aa9e8d",
        initial_quantity=1,
        name="test",
        royalties=1000,
        hash="abba",
        attributes=bytes("test", "utf-8"),
        uris=["a", "b"]
    )

    assert transaction.data
    assert transaction.data.decode() == "DCDTNFTCreate@4652414e4b2d616139653864@01@74657374@03e8@61626261@74657374@61@62"
    assert transaction.sender == grace.to_bech32()
    assert transaction.receiver == grace.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_setting_special_role_on_fungible_token():
    transaction = factory.create_transaction_for_setting_special_role_on_fungible_token(
        sender=frank,
        user=grace,
        token_identifier="FRANK-11ce3e",
        add_role_local_mint=True,
        add_role_local_burn=False,
    )

    assert transaction.data
    assert transaction.data.decode() == "setSpecialRole@4652414e4b2d313163653365@1e8a8b6b49de5b7be10aaa158a5a6a4abb4b56cc08f524bb5e6cd5f211ad3e13@44434454526f6c654c6f63616c4d696e74"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_setting_special_role_on_semi_fungible_token():
    transaction = factory.create_transaction_for_setting_special_role_on_semi_fungible_token(
        sender=frank,
        user=grace,
        token_identifier="FRANK-11ce3e",
        add_role_nft_create=True,
        add_role_nft_burn=True,
        add_role_nft_add_quantity=True,
        add_role_dcdt_transfer_role=True,
    )

    assert transaction.data
    assert transaction.data.decode() == "setSpecialRole@4652414e4b2d313163653365@1e8a8b6b49de5b7be10aaa158a5a6a4abb4b56cc08f524bb5e6cd5f211ad3e13@44434454526f6c654e4654437265617465@44434454526f6c654e46544275726e@44434454526f6c654e46544164645175616e74697479@444344545472616e73666572526f6c65"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_pausing():
    transaction = factory.create_transaction_for_pausing(
        sender=frank,
        token_identifier="FRANK-11ce3e",
    )

    assert transaction.data
    assert transaction.data.decode() == "pause@4652414e4b2d313163653365"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_unpausing():
    transaction = factory.create_transaction_for_unpausing(
        sender=frank,
        token_identifier="FRANK-11ce3e",
    )

    assert transaction.data
    assert transaction.data.decode() == "unPause@4652414e4b2d313163653365"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_freezing():
    transaction = factory.create_transaction_for_freezing(
        sender=frank,
        user=grace,
        token_identifier="FRANK-11ce3e",
    )

    assert transaction.data
    assert transaction.data.decode() == "freeze@4652414e4b2d313163653365@1e8a8b6b49de5b7be10aaa158a5a6a4abb4b56cc08f524bb5e6cd5f211ad3e13"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_unfreezing():
    transaction = factory.create_transaction_for_unfreezing(
        sender=frank,
        user=grace,
        token_identifier="FRANK-11ce3e",
    )

    assert transaction.data
    assert transaction.data.decode() == "unFreeze@4652414e4b2d313163653365@1e8a8b6b49de5b7be10aaa158a5a6a4abb4b56cc08f524bb5e6cd5f211ad3e13"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_local_minting():
    transaction = factory.create_transaction_for_local_minting(
        sender=frank,
        token_identifier="FRANK-11ce3e",
        supply_to_mint=10
    )

    assert transaction.data
    assert transaction.data.decode() == "DCDTLocalMint@4652414e4b2d313163653365@0a"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_local_burning():
    transaction = factory.create_transaction_for_local_burning(
        sender=frank,
        token_identifier="FRANK-11ce3e",
        supply_to_burn=10
    )

    assert transaction.data
    assert transaction.data.decode() == "DCDTLocalBurn@4652414e4b2d313163653365@0a"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_updating_attributes():
    transaction = factory.create_transaction_for_updating_attributes(
        sender=frank,
        token_identifier="FRANK-11ce3e",
        token_nonce=10,
        attributes=bytes("test", "utf-8"),
    )

    assert transaction.data
    assert transaction.data.decode() == "DCDTNFTUpdateAttributes@4652414e4b2d313163653365@0a@74657374"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_adding_quantity():
    transaction = factory.create_transaction_for_adding_quantity(
        sender=frank,
        token_identifier="FRANK-11ce3e",
        token_nonce=10,
        quantity_to_add=10
    )

    assert transaction.data
    assert transaction.data.decode() == "DCDTNFTAddQuantity@4652414e4b2d313163653365@0a@0a"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_burning_quantity():
    transaction = factory.create_transaction_for_burning_quantity(
        sender=frank,
        token_identifier="FRANK-11ce3e",
        token_nonce=10,
        quantity_to_burn=10
    )

    assert transaction.data
    assert transaction.data.decode() == "DCDTNFTBurn@4652414e4b2d313163653365@0a@0a"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_setting_burn_role_globally():
    transaction = factory.create_transaction_for_setting_burn_role_globally(
        sender=frank,
        token_identifier="FRANK-11ce3e",
    )

    assert transaction.data
    assert transaction.data.decode() == "setBurnRoleGlobally@4652414e4b2d313163653365"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_unsetting_burn_role_globally():
    transaction = factory.create_transaction_for_unsetting_burn_role_globally(
        sender=frank,
        token_identifier="FRANK-11ce3e",
    )

    assert transaction.data
    assert transaction.data.decode() == "unsetBurnRoleGlobally@4652414e4b2d313163653365"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0


def test_create_transaction_for_wiping():
    transaction = factory.create_transaction_for_wiping(
        sender=frank,
        user=grace,
        token_identifier="FRANK-11ce3e"
    )

    assert transaction.data
    assert transaction.data.decode() == "wipe@4652414e4b2d313163653365@1e8a8b6b49de5b7be10aaa158a5a6a4abb4b56cc08f524bb5e6cd5f211ad3e13"
    assert transaction.sender == frank.to_bech32()
    assert transaction.value == 0




from drt_sdk_core.constants import INTEGER_MAX_NUM_BYTES


def encode_unsigned_number(arg: int) -> bytes:
    return arg.to_bytes(INTEGER_MAX_NUM_BYTES, byteorder="big", signed=False).lstrip(bytes([0]))


def encode_signed_number(arg: int) -> bytes:
    if arg == 0:
        return b''
    length = ((arg + (arg < 0)).bit_length() + 7 + 1) // 8
    return arg.to_bytes(length, byteorder="big", signed=True)


def decode_unsigned_number(arg: bytes) -> int:
    return int.from_bytes(arg, byteorder="big", signed=False)


def decode_signed_number(arg: bytes) -> int:
    return int.from_bytes(arg, byteorder="big", signed=True)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Transaction(_message.Message):
    __slots__ = ["ChainID", "Data", "GasLimit", "GasPrice", "GuardAddr", "GuardSignature", "Nonce", "Options", "RcvAddr", "RcvUserName", "Signature", "SndAddr", "SndUserName", "Value", "Version"]
    CHAINID_FIELD_NUMBER: _ClassVar[int]
    ChainID: bytes
    DATA_FIELD_NUMBER: _ClassVar[int]
    Data: bytes
    GASLIMIT_FIELD_NUMBER: _ClassVar[int]
    GASPRICE_FIELD_NUMBER: _ClassVar[int]
    GUARDADDR_FIELD_NUMBER: _ClassVar[int]
    GUARDSIGNATURE_FIELD_NUMBER: _ClassVar[int]
    GasLimit: int
    GasPrice: int
    GuardAddr: bytes
    GuardSignature: bytes
    NONCE_FIELD_NUMBER: _ClassVar[int]
    Nonce: int
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    Options: int
    RCVADDR_FIELD_NUMBER: _ClassVar[int]
    RCVUSERNAME_FIELD_NUMBER: _ClassVar[int]
    RcvAddr: bytes
    RcvUserName: bytes
    SIGNATURE_FIELD_NUMBER: _ClassVar[int]
    SNDADDR_FIELD_NUMBER: _ClassVar[int]
    SNDUSERNAME_FIELD_NUMBER: _ClassVar[int]
    Signature: bytes
    SndAddr: bytes
    SndUserName: bytes
    VALUE_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    Value: bytes
    Version: int
    def __init__(self, Nonce: _Optional[int] = ..., Value: _Optional[bytes] = ..., RcvAddr: _Optional[bytes] = ..., RcvUserName: _Optional[bytes] = ..., SndAddr: _Optional[bytes] = ..., SndUserName: _Optional[bytes] = ..., GasPrice: _Optional[int] = ..., GasLimit: _Optional[int] = ..., Data: _Optional[bytes] = ..., ChainID: _Optional[bytes] = ..., Version: _Optional[int] = ..., Signature: _Optional[bytes] = ..., Options: _Optional[int] = ..., GuardAddr: _Optional[bytes] = ..., GuardSignature: _Optional[bytes] = ...) -> None: ...



from enum import Enum


class ByteZero(Enum):
    Upgradeable = 1
    Reserved2 = 2
    Readable = 4


class ByteOne(Enum):
    Reserved1 = 1
    Payable = 2
    PayableByContract = 4


class CodeMetadata:
    def __init__(self, upgradeable: bool = True, readable: bool = True, payable: bool = False, payable_by_contract: bool = False):
        self.upgradeable = upgradeable
        self.readable = readable
        self.payable = payable
        self.payable_by_contract = payable_by_contract

    def serialize(self) -> bytes:
        data = bytearray([0, 0])

        if self.upgradeable:
            data[0] |= ByteZero.Upgradeable.value
        if self.readable:
            data[0] |= ByteZero.Readable.value
        if self.payable:
            data[1] |= ByteOne.Payable.value
        if self.payable_by_contract:
            data[1] |= ByteOne.PayableByContract.value

        return bytes(data)

    def __str__(self) -> str:
        return self.serialize().hex()




import pytest

from drt_sdk_core.address import (Address, AddressComputer,
                                         AddressFactory, is_valid_bech32)
from drt_sdk_core.errors import ErrBadAddress, ErrBadPubkeyLength


def test_address():
    address = Address.new_from_bech32("moa1l453hd0gt5gzdp7czpuall8ggt2dcv5zwmfdf3sd3lguxseux2fskgws3j")
    assert "fd691bb5e85d102687d81079dffce842d4dc328276d2d4c60d8fd1c3433c3293" == address.to_hex()
    assert "moa1l453hd0gt5gzdp7czpuall8ggt2dcv5zwmfdf3sd3lguxseux2fskgws3j" == address.to_bech32()

    address = Address.new_from_hex("fd691bb5e85d102687d81079dffce842d4dc328276d2d4c60d8fd1c3433c3293", "moa")
    assert "fd691bb5e85d102687d81079dffce842d4dc328276d2d4c60d8fd1c3433c3293" == address.to_hex()
    assert "moa1l453hd0gt5gzdp7czpuall8ggt2dcv5zwmfdf3sd3lguxseux2fskgws3j" == address.to_bech32()

    with pytest.raises(ErrBadPubkeyLength):
        address = Address(bytes(), "moa")

    with pytest.raises(ErrBadAddress):
        address = Address.new_from_bech32("bad")


def test_address_with_custom_hrp():
    address = Address.new_from_hex("0139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1", "test")
    assert address.hrp == "test"
    assert address.to_bech32() == "test1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ss5hqhtr"

    address = Address.new_from_bech32("test1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ss5hqhtr")
    assert address.hrp == "test"
    assert address.to_hex() == "0139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1"


def test_address_factory():
    factory_foo = AddressFactory("foo")
    factory_moa = AddressFactory("moa")
    pubkey = bytes.fromhex("0139472eff6886771a982f3083da5d421f24c29181e63888228dc81ca60d69e1")

    assert factory_foo.create_from_public_key(pubkey).to_bech32() == "foo1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssunhpj4"
    assert factory_moa.create_from_public_key(pubkey).to_bech32() == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"


def test_is_valid_bech32():
    assert is_valid_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8", "moa")
    assert is_valid_bech32("foo1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssunhpj4", "foo")
    assert not is_valid_bech32("foobar", "foo")
    assert not is_valid_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8", "foo")


def test_get_address_shard():
    address_computer = AddressComputer()
    address = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
    assert address_computer.get_shard_of_address(address) == 1

    address = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")
    assert address_computer.get_shard_of_address(address) == 0

    address = Address.new_from_bech32("moa1k2s324ww2g0yj38qn2ch2jwctdy8mnfxep94q9arncc6xecg3xaqhr5l9h")
    assert address_computer.get_shard_of_address(address) == 2


def test_compute_contract_address():
    deployer = Address.new_from_bech32("moa1j0hxzs7dcyxw08c4k2nv9tfcaxmqy8rj59meq505w92064x0h40q4737p3")
    address_computer = AddressComputer()

    contract_address = address_computer.compute_contract_address(deployer, deployment_nonce=0)
    assert contract_address.to_hex() == "00000000000000000500bb652200ed1f994200ab6699462cab4b1af7b11ebd5e"
    assert contract_address.to_bech32() == "moa1qqqqqqqqqqqqqpgqhdjjyq8dr7v5yq9tv6v5vt9tfvd00vg7h40qhxc27r"

    contract_address = address_computer.compute_contract_address(deployer, deployment_nonce=1)
    assert contract_address.to_hex() == "000000000000000005006e4f90488e27342f9a46e1809452c85ee7186566bd5e"
    assert contract_address.to_bech32() == "moa1qqqqqqqqqqqqqpgqde8eqjywyu6zlxjxuxqfg5kgtmn3setxh40q5tpk55"



import base64
import json
from typing import Any, Dict, List, Protocol

from drt_sdk_core import Address
from drt_sdk_core.errors import InvalidInnerTransactionError
from drt_sdk_core.interfaces import IAddress, ITransaction
from drt_sdk_core.serializer import args_to_string
from drt_sdk_core.transaction import Transaction


class IConfig(Protocol):
    chain_id: str
    min_gas_limit: int
    gas_limit_per_byte: int


class RelayedTransactionsFactory:
    def __init__(self, config: IConfig) -> None:
        self._config = config

    def create_relayed_v1_transaction(self,
                                      inner_transaction: ITransaction,
                                      relayer_address: IAddress) -> Transaction:
        if not inner_transaction.gas_limit:
            raise InvalidInnerTransactionError("The gas limit is not set for the inner transaction")

        if not inner_transaction.signature:
            raise InvalidInnerTransactionError("The inner transaction is not signed")

        serialized_transaction = self._prepare_inner_transaction_for_relayed_v1(inner_transaction)
        data = f"relayedTx@{serialized_transaction.encode().hex()}"

        gas_limit = self._config.min_gas_limit + self._config.gas_limit_per_byte * len(data) + inner_transaction.gas_limit

        return Transaction(
            chain_id=self._config.chain_id,
            sender=relayer_address.to_bech32(),
            receiver=inner_transaction.sender,
            gas_limit=gas_limit,
            data=data.encode()
        )

    def create_relayed_v2_transaction(self,
                                      inner_transaction: ITransaction,
                                      inner_transaction_gas_limit: int,
                                      relayer_address: IAddress) -> Transaction:
        if inner_transaction.gas_limit:
            raise InvalidInnerTransactionError("The gas limit should not be set for the inner transaction")

        if not inner_transaction.signature:
            raise InvalidInnerTransactionError("The inner transaction is not signed")

        arguments: List[Any] = [
            Address.new_from_bech32(inner_transaction.receiver),
            inner_transaction.nonce,
            inner_transaction.data,
            inner_transaction.signature
        ]

        data = f"relayedTxV2@{args_to_string(arguments)}"
        gas_limit = inner_transaction_gas_limit + self._config.min_gas_limit + self._config.gas_limit_per_byte * len(data)

        return Transaction(
            sender=relayer_address.to_bech32(),
            receiver=inner_transaction.sender,
            value=0,
            gas_limit=gas_limit,
            chain_id=self._config.chain_id,
            data=data.encode(),
            version=inner_transaction.version,
            options=inner_transaction.options
        )

    def _prepare_inner_transaction_for_relayed_v1(self, inner_transaction: ITransaction) -> str:
        sender = Address.new_from_bech32(inner_transaction.sender).to_hex()
        receiver = Address.new_from_bech32(inner_transaction.receiver).to_hex()

        tx: Dict[str, Any] = {
            "nonce": inner_transaction.nonce,
            "sender": base64.b64encode(bytes.fromhex(sender)).decode(),
            "receiver": base64.b64encode(bytes.fromhex(receiver)).decode(),
            "value": inner_transaction.value,
            "gasPrice": inner_transaction.gas_price,
            "gasLimit": inner_transaction.gas_limit,
            "data": base64.b64encode(inner_transaction.data).decode(),
            "signature": base64.b64encode(inner_transaction.signature).decode(),
            "chainID": base64.b64encode(inner_transaction.chain_id.encode()).decode(),
            "version": inner_transaction.version,
        }

        if inner_transaction.options:
            tx["options"] = inner_transaction.options

        if inner_transaction.guardian:
            guardian = Address.new_from_bech32(inner_transaction.guardian).to_hex()
            tx["guardian"] = base64.b64encode(bytes.fromhex(guardian)).decode()

        if inner_transaction.guardian_signature:
            tx["guardianSignature"] = base64.b64encode(inner_transaction.guardian_signature).decode()

        if inner_transaction.sender_username:
            tx["sndUserName"] = base64.b64encode(inner_transaction.sender_username.encode()).decode()

        if inner_transaction.receiver_username:
            tx[f"rcvUserName"] = base64.b64encode(inner_transaction.receiver_username.encode()).decode()

        return json.dumps(tx, separators=(",", ":"))



import json
from base64 import b64encode
from collections import OrderedDict
from hashlib import blake2b
from typing import Any, Dict, Optional, Protocol

from Cryptodome.Hash import keccak

from drt_sdk_core.constants import (DEFAULT_HRP, DIGEST_SIZE,
                                           TRANSACTION_MIN_GAS_PRICE,
                                           TRANSACTION_OPTIONS_DEFAULT,
                                           TRANSACTION_VERSION_DEFAULT)
from drt_sdk_core.errors import NotEnoughGasError
from drt_sdk_core.interfaces import INetworkConfig, ITransaction
from drt_sdk_core.proto.transaction_serializer import ProtoSerializer


class IAddressConverter(Protocol):
    def __init__(self, hrp: str = DEFAULT_HRP) -> None:
        ...

    def bech32_to_pubkey(self, value: str) -> bytes:
        ...

    def pubkey_to_bech32(self, pubkey: bytes) -> str:
        ...


class Transaction:
    def __init__(self,
                 sender: str,
                 receiver: str,
                 gas_limit: int,
                 chain_id: str,
                 nonce: Optional[int] = None,
                 value: Optional[int] = None,
                 sender_username: Optional[str] = None,
                 receiver_username: Optional[str] = None,
                 gas_price: Optional[int] = None,
                 data: Optional[bytes] = None,
                 version: Optional[int] = None,
                 options: Optional[int] = None,
                 guardian: Optional[str] = None,
                 signature: Optional[bytes] = None,
                 guardian_signature: Optional[bytes] = None
                 ) -> None:
        self.chain_id = chain_id
        self.sender = sender
        self.receiver = receiver
        self.gas_limit = gas_limit

        self.nonce = nonce or 0
        self.value = value or 0
        self.data = data or bytes()
        self.signature = signature or bytes()

        self.sender_username = sender_username or ""
        self.receiver_username = receiver_username or ""

        self.gas_price = gas_price or TRANSACTION_MIN_GAS_PRICE
        self.version = version or TRANSACTION_VERSION_DEFAULT
        self.options = options or TRANSACTION_OPTIONS_DEFAULT

        self.guardian = guardian or ""
        self.guardian_signature = guardian_signature or bytes()


class TransactionComputer:
    def __init__(self) -> None:
        pass

    def compute_transaction_fee(self, transaction: ITransaction, network_config: INetworkConfig) -> int:
        move_balance_gas = network_config.min_gas_limit + len(transaction.data) * network_config.gas_per_data_byte
        if move_balance_gas > transaction.gas_limit:
            raise NotEnoughGasError(transaction.gas_limit)

        fee_for_move = move_balance_gas * transaction.gas_price
        if move_balance_gas == transaction.gas_limit:
            return int(fee_for_move)

        diff = transaction.gas_limit - move_balance_gas
        modified_gas_price = transaction.gas_price * network_config.gas_price_modifier
        processing_fee = diff * modified_gas_price

        return int(fee_for_move + processing_fee)

    def compute_bytes_for_signing(self, transaction: ITransaction) -> bytes:
        dictionary = self._to_dictionary(transaction)
        serialized = self._dict_to_json(dictionary)
        return serialized

    def compute_hash_for_signing(self, transaction: ITransaction) -> bytes:
        return keccak.new(digest_bits=256).update(self.compute_bytes_for_signing(transaction)).digest()

    def compute_transaction_hash(self, transaction: ITransaction) -> bytes:
        proto = ProtoSerializer()
        serialized_tx = proto.serialize_transaction(transaction)
        tx_hash = blake2b(serialized_tx, digest_size=DIGEST_SIZE).hexdigest()
        return bytes.fromhex(tx_hash)

    def _to_dictionary(self, transaction: ITransaction) -> Dict[str, Any]:
        dictionary: Dict[str, Any] = OrderedDict()
        dictionary["nonce"] = transaction.nonce
        dictionary["value"] = str(transaction.value)

        dictionary["receiver"] = transaction.receiver
        dictionary["sender"] = transaction.sender

        if transaction.sender_username:
            dictionary["senderUsername"] = b64encode(transaction.sender_username.encode()).decode()

        if transaction.receiver_username:
            dictionary["receiverUsername"] = b64encode(transaction.receiver_username.encode()).decode()

        dictionary["gasPrice"] = transaction.gas_price
        dictionary["gasLimit"] = transaction.gas_limit

        if transaction.data:
            dictionary["data"] = b64encode(transaction.data).decode()

        dictionary["chainID"] = transaction.chain_id

        if transaction.version:
            dictionary["version"] = transaction.version

        if transaction.options:
            dictionary["options"] = transaction.options

        if transaction.guardian:
            dictionary["guardian"] = transaction.guardian

        return dictionary

    def _dict_to_json(self, dictionary: Dict[str, Any]) -> bytes:
        serialized = json.dumps(dictionary, separators=(',', ':')).encode("utf-8")
        return serialized



from drt_sdk_core.address import Address
from drt_sdk_core.tokens import Token, TokenComputer, TokenTransfer
from drt_sdk_core.transaction_factories.transactions_factory_config import \
    TransactionsFactoryConfig
from drt_sdk_core.transaction_factories.transfer_transactions_factory import \
    TransferTransactionsFactory


class TestTransferTransactionsFactory:
    transfer_factory = TransferTransactionsFactory(TransactionsFactoryConfig("D"), TokenComputer())

    def test_create_transaction_for_native_token_transfer_no_data(self):
        alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")

        transaction = self.transfer_factory.create_transaction_for_native_token_transfer(
            sender=alice,
            receiver=bob,
            native_amount=1000000000000000000
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk"
        assert transaction.value == 1000000000000000000
        assert transaction.chain_id == "D"
        assert transaction.gas_limit == 50_000
        assert transaction.data == b""

    def test_create_transaction_for_native_token_transfer_with_data(self):
        alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")

        transaction = self.transfer_factory.create_transaction_for_native_token_transfer(
            sender=alice,
            receiver=bob,
            native_amount=1000000000000000000,
            data="test data"
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk"
        assert transaction.value == 1000000000000000000
        assert transaction.chain_id == "D"
        assert transaction.gas_limit == 63_500
        assert transaction.data == b"test data"

    def test_create_transaction_for_dcdt_transfer(self):
        alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")

        foo_token = Token("FOO-123456")
        token_transfer = TokenTransfer(foo_token, 1000000)

        transaction = self.transfer_factory.create_transaction_for_dcdt_token_transfer(
            sender=alice,
            receiver=bob,
            token_transfers=[token_transfer]
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk"
        assert transaction.value == 0
        assert transaction.chain_id == "D"
        assert transaction.data.decode() == "DCDTTransfer@464f4f2d313233343536@0f4240"
        assert transaction.gas_limit == 410_000

    def test_create_transaction_for_nft_transfer(self):
        alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")

        nft = Token("NFT-123456", 10)
        token_transfer = TokenTransfer(nft, 1)

        transaction = self.transfer_factory.create_transaction_for_dcdt_token_transfer(
            sender=alice,
            receiver=bob,
            token_transfers=[token_transfer]
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.value == 0
        assert transaction.chain_id == "D"
        assert transaction.data.decode() == "DCDTNFTTransfer@4e46542d313233343536@0a@01@8049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f8"
        assert transaction.gas_limit == 1_210_500

    def test_create_transaction_for_multiple_nft_transfers(self):
        alice = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")
        bob = Address.new_from_bech32("moa1spyavw0956vq68xj8y4tenjpq2wd5a9p2c6j8gsz7ztyrnpxrruq0yu4wk")

        first_nft = Token("NFT-123456", 10)
        first_transfer = TokenTransfer(first_nft, 1)

        second_nft = Token("TEST-987654", 1)
        second_transfer = TokenTransfer(second_nft, 1)

        transaction = self.transfer_factory.create_transaction_for_dcdt_token_transfer(
            sender=alice,
            receiver=bob,
            token_transfers=[first_transfer, second_transfer]
        )

        assert transaction.sender == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.receiver == "moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8"
        assert transaction.value == 0
        assert transaction.chain_id == "D"
        assert transaction.data.decode() == "MultiDCDTNFTTransfer@8049d639e5a6980d1cd2392abcce41029cda74a1563523a202f09641cc2618f8@02@4e46542d313233343536@0a@01@544553542d393837363534@01@01"
        assert transaction.gas_limit == 1_466_000




from drt_sdk_core.address import Address
from drt_sdk_core.transaction_builders.default_configuration import \
    DefaultTransactionBuildersConfiguration
from drt_sdk_core.transaction_builders.dcdt_builders import \
    DCDTIssueBuilder

dummyConfig = DefaultTransactionBuildersConfiguration(chain_id="D")


def test_dcdt_issue_builder():
    issuer = Address.new_from_bech32("moa1qyu5wthldzr8wx5c9ucg8kjagg0jfs53s8nr3zpz3hypefsdd8ssfq94h8")

    builder = DCDTIssueBuilder(
        config=dummyConfig,
        issuer=issuer,
        token_name="FOO",
        token_ticker="FOO",
        initial_supply=1000000000000,
        num_decimals=8,
        can_freeze=True,
        can_mint=True,
        can_upgrade=True
    )

    payload = builder.build_payload()
    tx = builder.build()

    assert payload.data == b"issue@464f4f@464f4f@e8d4a51000@08@63616e467265657a65@74727565@63616e4d696e74@74727565@63616e55706772616465@74727565@63616e4164645370656369616c526f6c6573@66616c7365"
    assert tx.chain_id == "D"
    assert tx.sender == issuer.to_bech32()
    assert tx.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzllls29jpxv"
    assert tx.gas_limit == 50000 + payload.length() * 1500 + 60000000
    assert tx.data.decode() == str(payload)





from drt_sdk_core.code_metadata import CodeMetadata


def test_code_metadata_serialize():
    assert CodeMetadata().serialize() == bytes([0x05, 0x00])
    assert CodeMetadata(upgradeable=True, readable=True).serialize() == bytes([0x05, 0x00])
    assert CodeMetadata(upgradeable=True, readable=True, payable=True, payable_by_contract=True).serialize() == bytes([0x05, 0x06])
    assert CodeMetadata(upgradeable=True, readable=True, payable=False, payable_by_contract=True).serialize() == bytes([0x05, 0x04])
    assert CodeMetadata(upgradeable=False, readable=False, payable=False, payable_by_contract=False).serialize() == bytes([0x00, 0x00])




from dataclasses import dataclass
from typing import List


@dataclass
class DCDTIssueOutcome:
    token_identifier: str


@dataclass
class RegisterAndSetAllRolesOutcome:
    token_identifier: str
    roles: List[str]


@dataclass
class SetSpecialRoleOutcome:
    user_address: str
    token_identifier: str
    roles: List[str]


@dataclass
class NFTCreateOutcome:
    token_identifier: str
    nonce: int
    initial_quantity: int


@dataclass
class MintOutcome:
    user_address: str
    token_identifier: str
    nonce: int
    minted_supply: int


@dataclass
class BurnOutcome:
    user_address: str
    token_identifier: str
    nonce: int
    burnt_supply: int


@dataclass
class PausingOutcome:
    pass


@dataclass
class FreezingOutcome:
    user_address: str
    token_identifier: str
    nonce: int
    balance: int


@dataclass
class WipingOutcome:
    user_address: str
    token_identifier: str
    nonce: int
    balance: int


@dataclass
class UpdateAttributesOutcome:
    token_identifier: str
    nonce: int
    attributes: bytes


@dataclass
class AddQuantityOutcome:
    token_identifier: str
    nonce: int
    added_quantity: int


@dataclass
class BurnQuantityOutcome:
    token_identifier: str
    nonce: int
    burnt_quantity: int




from decimal import Decimal
from typing import Any

import pytest

from drt_sdk_core.token_payment import TokenPayment


def test_with_rewa():
    assert str(TokenPayment.rewa_from_amount(Decimal("1"))) == "1000000000000000000"
    assert str(TokenPayment.rewa_from_amount(Decimal("10"))) == "10000000000000000000"
    assert str(TokenPayment.rewa_from_amount(Decimal("100"))) == "100000000000000000000"
    assert str(TokenPayment.rewa_from_amount(Decimal("1000"))) == "1000000000000000000000"
    assert str(TokenPayment.rewa_from_amount(Decimal("0.1"))) == "100000000000000000"
    assert str(TokenPayment.rewa_from_amount(Decimal("0.123456789"))) == "123456789000000000"
    assert str(TokenPayment.rewa_from_amount(Decimal("0.123456789123456789"))) == "123456789123456789"
    assert str(TokenPayment.rewa_from_amount(Decimal("0.123456789123456789777"))) == "123456789123456789"
    assert str(TokenPayment.rewa_from_amount(Decimal("0.123456789123456789777777888888"))) == "123456789123456789"

    assert TokenPayment.rewa_from_amount(Decimal("0.1")).to_amount_string() == "0.100000000000000000"
    assert TokenPayment.rewa_from_amount(Decimal("1")).to_amount_string() == "1.000000000000000000"
    assert TokenPayment.rewa_from_amount(Decimal("10")).to_amount_string() == "10.000000000000000000"
    assert TokenPayment.rewa_from_amount(Decimal("100.00")).to_amount_string() == "100.000000000000000000"
    assert TokenPayment.rewa_from_amount(Decimal("1000")).to_amount_string() == "1000.000000000000000000"
    assert TokenPayment.rewa_from_amount(Decimal("1000")).to_amount_string(normalize=True) == "1000"
    assert TokenPayment.rewa_from_amount(Decimal("0.123456789")).to_amount_string() == "0.123456789000000000"
    assert TokenPayment.rewa_from_amount(Decimal("0.123456789")).to_amount_string(normalize=True) == "0.123456789"
    assert TokenPayment.rewa_from_amount(Decimal("0.123456789123456789777777888888")).to_amount_string() == "0.123456789123456789"

    assert str(TokenPayment.rewa_from_integer(0)) == "0"
    assert TokenPayment.rewa_from_integer(0).to_amount_string() == "0.000000000000000000"

    assert str(TokenPayment.rewa_from_integer(1)) == "1"
    assert TokenPayment.rewa_from_integer(1).to_amount_string() == "0.000000000000000001"

    assert str(TokenPayment.rewa_from_integer(123456789123456789)) == "123456789123456789"
    assert TokenPayment.rewa_from_integer(123456789123456789).to_amount_string() == "0.123456789123456789"


def test_with_fungible_dcdt():
    identifier = "USDC-c76f1f"
    num_decimals = 6

    assert str(TokenPayment.fungible_from_amount(identifier, Decimal("1"), num_decimals)) == "1000000"
    assert str(TokenPayment.fungible_from_amount(identifier, Decimal("0.1"), num_decimals)) == "100000"
    assert str(TokenPayment.fungible_from_amount(identifier, Decimal("0.123456789"), num_decimals)) == "123456"
    assert str(TokenPayment.fungible_from_integer(identifier, 1000000, num_decimals)) == "1000000"
    assert TokenPayment.fungible_from_integer(identifier, 1000000, num_decimals).to_amount_string() == "1.000000"


def test_with_meta_dcdt():
    identifier = "MEXFARML-28d646"
    num_decimals = 18
    nonce = 12345678

    payment = TokenPayment.meta_dcdt_from_amount(identifier, nonce, Decimal("0.1"), num_decimals)
    assert str(payment) == "100000000000000000"
    assert payment.to_amount_string() == "0.100000000000000000"
    assert payment.token_identifier == identifier
    assert payment.token_nonce == nonce


def test_with_nft():
    identifier = "TEST-38f249"
    nonce = 1

    payment = TokenPayment.non_fungible(identifier, nonce)
    assert str(payment) == "1"
    assert payment.to_amount_string() == "1"
    assert payment.token_identifier == identifier
    assert payment.token_nonce == nonce


def test_misuse_raises_error():
    with pytest.raises(ValueError, match="amount_as_integer must be an integer"):
        amount: Any = "1"
        TokenPayment.rewa_from_integer(amount)

    with pytest.raises(ValueError, match="amount_as_integer must be an integer"):
        amount: Any = "1"
        TokenPayment.fungible_from_integer("USDC-c76f1f", amount, 6)



from decimal import ROUND_DOWN, Context, Decimal, localcontext
from typing import Union

from drt_sdk_core import typecheck
from drt_sdk_core.constants import (REWA_NUM_DECIMALS,
                                           REWA_TOKEN_IDENTIFIER)
from drt_sdk_core.interfaces import INonce, ITokenIdentifier


class TokenPayment:
    def __init__(self, token_identifier: ITokenIdentifier, token_nonce: INonce, amount_as_integer: int, num_decimals: int) -> None:
        typecheck.assert_is_integer(amount_as_integer, "amount_as_integer must be an integer")

        self.token_identifier: ITokenIdentifier = token_identifier
        self.token_nonce: INonce = token_nonce
        self.amount_as_integer: int = amount_as_integer
        self.num_decimals = num_decimals

    def is_rewa(self):
        return self.token_identifier == REWA_TOKEN_IDENTIFIER

    def is_fungible(self):
        return self.token_nonce == 0

    @classmethod
    def rewa_from_amount(cls, amount: Union[Decimal, str, int]) -> 'TokenPayment':
        amount = Decimal(amount)
        amount_as_integer = cls._amount_to_integer(amount, REWA_NUM_DECIMALS)
        return cls.rewa_from_integer(amount_as_integer)

    @classmethod
    def rewa_from_integer(cls, amount_as_integer: int) -> 'TokenPayment':
        return cls(REWA_TOKEN_IDENTIFIER, 0, amount_as_integer, REWA_NUM_DECIMALS)

    @classmethod
    def fungible_from_amount(cls, token_identifier: ITokenIdentifier, amount: Union[Decimal, str, int], num_decimals: int) -> 'TokenPayment':
        amount = Decimal(amount)
        amount_as_integer = cls._amount_to_integer(amount, num_decimals)
        return cls.fungible_from_integer(token_identifier, amount_as_integer, num_decimals)

    @classmethod
    def fungible_from_integer(cls, token_identifier: ITokenIdentifier, amount_as_integer: int, num_decimals: int) -> 'TokenPayment':
        return cls(token_identifier, 0, amount_as_integer, num_decimals)

    @classmethod
    def non_fungible(cls, token_identifier: ITokenIdentifier, nonce: INonce) -> 'TokenPayment':
        return cls(token_identifier, nonce, 1, 0)

    @classmethod
    def semi_fungible(cls, token_identifier: ITokenIdentifier, nonce: INonce, quantity: int) -> 'TokenPayment':
        return cls(token_identifier, nonce, quantity, 0)

    @classmethod
    def meta_dcdt_from_amount(cls, token_identifier: ITokenIdentifier, nonce: int, amount: Union[Decimal, str, int], num_decimals: int) -> 'TokenPayment':
        amount = Decimal(amount)
        amount_as_integer = cls._amount_to_integer(amount, num_decimals)
        return cls.meta_dcdt_from_integer(token_identifier, nonce, amount_as_integer, num_decimals)

    @classmethod
    def meta_dcdt_from_integer(cls, token_identifier: ITokenIdentifier, nonce: int, amount_as_integer: int, num_decimals: int) -> 'TokenPayment':
        return cls(token_identifier, nonce, amount_as_integer, num_decimals)

    def to_amount_string(self, normalize: bool = False) -> str:
        with localcontext() as ctx:
            self._adjust_decimal_context(ctx)
            amount = Decimal(self.amount_as_integer).scaleb(-self.num_decimals)

            if normalize:
                amount = amount.normalize()

            return f"{amount:f}"

    @classmethod
    def _amount_to_integer(cls, amount: Decimal, num_decimals: int) -> int:
        with localcontext() as ctx:
            cls._adjust_decimal_context(ctx)
            amount_as_integer = int(amount.scaleb(num_decimals))
            return amount_as_integer

    @ classmethod
    def _adjust_decimal_context(cls, context: Context):
        context.prec = 128
        context.rounding = ROUND_DOWN

    def __str__(self) -> str:
        return str(self.amount_as_integer)

    def __repr__(self) -> str:
        return str(self.amount_as_integer)



from drt_sdk_wallet import ValidatorSecretKey, ValidatorSigner

from drt_sdk_core.address import Address
from drt_sdk_core.constants import DELEGATION_MANAGER_SC_ADDRESS
from drt_sdk_core.transaction_factories.delegation_transactions_factory import \
    DelegationTransactionsFactory
from drt_sdk_core.transaction_factories.transactions_factory_config import \
    TransactionsFactoryConfig


class TestDelegationTransactionsFactory:
    config = TransactionsFactoryConfig("D")
    factory = DelegationTransactionsFactory(config)

    def test_create_transaction_for_new_delegation_contract(self):
        transaction = self.factory.create_transaction_for_new_delegation_contract(
            sender=Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"),
            total_delegation_cap=5000000000000000000000,
            service_fee=10,
            amount=1250000000000000000000
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == DELEGATION_MANAGER_SC_ADDRESS
        assert transaction.data
        assert transaction.data.decode() == "createNewDelegationContract@010f0cf064dd59200000@0a"
        assert transaction.gas_limit == 60126500
        assert transaction.value == 1250000000000000000000

    def test_create_transaction_for_adding_nodes(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        validator_secret_key = ValidatorSecretKey.from_string("7cff99bd671502db7d15bc8abc0c9a804fb925406fbdd50f1e4c17a4cd774247")
        validator_signer = ValidatorSigner(validator_secret_key)

        signed_message = validator_signer.sign(bytes.fromhex(delegation_contract.to_hex()))
        public_key = validator_secret_key.generate_public_key()

        public_keys = [public_key]
        signed_messages = [signed_message]

        transaction = self.factory.create_transaction_for_adding_nodes(
            sender=sender,
            delegation_contract=delegation_contract,
            public_keys=public_keys,
            signed_messages=signed_messages
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "addNodes@e7beaa95b3877f47348df4dd1cb578a4f7cabf7a20bfeefe5cdd263878ff132b765e04fef6f40c93512b666c47ed7719b8902f6c922c04247989b7137e837cc81a62e54712471c97a2ddab75aa9c2f58f813ed4c0fa722bde0ab718bff382208@81109fa1c8d3dc7b6c2d6e65206cc0bc1a83c9b2d1eb91a601d66ad32def430827d5eb52917bd2b0d04ce195738db216"
        assert transaction.value == 0

    def test_create_transaction_for_removing_nodes(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        public_keys = ["notavalidblskeyhexencoded".encode()]

        transaction = self.factory.create_transaction_for_removing_nodes(
            sender=sender,
            delegation_contract=delegation_contract,
            public_keys=public_keys
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "removeNodes@6e6f746176616c6964626c736b6579686578656e636f646564"
        assert transaction.value == 0

    def test_create_transaction_for_staking_nodes(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        public_keys = ["notavalidblskeyhexencoded".encode()]

        transaction = self.factory.create_transaction_for_staking_nodes(
            sender=sender,
            delegation_contract=delegation_contract,
            public_keys=public_keys
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "stakeNodes@6e6f746176616c6964626c736b6579686578656e636f646564"
        assert transaction.value == 0

    def test_create_transaction_for_unbonding_nodes(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        public_keys = ["notavalidblskeyhexencoded".encode()]

        transaction = self.factory.create_transaction_for_unbonding_nodes(
            sender=sender,
            delegation_contract=delegation_contract,
            public_keys=public_keys
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "unBondNodes@6e6f746176616c6964626c736b6579686578656e636f646564"
        assert transaction.value == 0

    def test_create_transaction_for_unstaking_nodes(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        public_keys = ["notavalidblskeyhexencoded".encode()]

        transaction = self.factory.create_transaction_for_unstaking_nodes(
            sender=sender,
            delegation_contract=delegation_contract,
            public_keys=public_keys
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "unStakeNodes@6e6f746176616c6964626c736b6579686578656e636f646564"
        assert transaction.value == 0

    def test_create_transaction_for_unjailing_nodes(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        public_keys = ["notavalidblskeyhexencoded".encode()]

        transaction = self.factory.create_transaction_for_unjailing_nodes(
            sender=sender,
            delegation_contract=delegation_contract,
            public_keys=public_keys
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "unJailNodes@6e6f746176616c6964626c736b6579686578656e636f646564"
        assert transaction.value == 0

    def test_create_transaction_for_changing_service_fee(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_changing_service_fee(
            sender=sender,
            delegation_contract=delegation_contract,
            service_fee=10
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "changeServiceFee@0a"
        assert transaction.value == 0

    def test_create_transaction_for_modifying_delegation_cap(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_modifying_delegation_cap(
            sender=sender,
            delegation_contract=delegation_contract,
            delegation_cap=5000000000000000000000
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "modifyTotalDelegationCap@010f0cf064dd59200000"
        assert transaction.value == 0

    def test_create_transaction_for_setting_automatic_activation(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_setting_automatic_activation(
            sender=sender,
            delegation_contract=delegation_contract
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "setAutomaticActivation@74727565"
        assert transaction.value == 0

    def test_create_transaction_for_unsetting_automatic_activation(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_unsetting_automatic_activation(
            sender=sender,
            delegation_contract=delegation_contract
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "setAutomaticActivation@66616c7365"
        assert transaction.value == 0

    def test_create_transaction_for_setting_cap_check_on_redelegate_rewards(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_setting_cap_check_on_redelegate_rewards(
            sender=sender,
            delegation_contract=delegation_contract
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "setCheckCapOnReDelegateRewards@74727565"
        assert transaction.value == 0

    def test_create_transaction_for_unsetting_cap_check_on_redelegate_rewards(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_unsetting_cap_check_on_redelegate_rewards(
            sender=sender,
            delegation_contract=delegation_contract
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "setCheckCapOnReDelegateRewards@66616c7365"
        assert transaction.value == 0

    def test_create_transaction_for_setting_metadata(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_setting_metadata(
            sender=sender,
            delegation_contract=delegation_contract,
            name="name",
            website="website",
            identifier="identifier"
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "setMetaData@6e616d65@77656273697465@6964656e746966696572"
        assert transaction.value == 0

    def test_create_transaction_for_delegating(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_delegating(
            sender=sender,
            delegation_contract=delegation_contract,
            amount=1000000000000000000
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "delegate"
        assert transaction.value == 1000000000000000000
        assert transaction.gas_limit == 12000000

    def test_create_transaction_for_claiming_rewards(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_claiming_rewards(
            sender=sender,
            delegation_contract=delegation_contract
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "claimRewards"
        assert transaction.value == 0
        assert transaction.gas_limit == 6000000

    def test_create_transaction_for_redelegating_rewards(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_redelegating_rewards(
            sender=sender,
            delegation_contract=delegation_contract
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "reDelegateRewards"
        assert transaction.value == 0
        assert transaction.gas_limit == 12000000

    def test_create_transaction_for_undelegating(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_undelegating(
            sender=sender,
            delegation_contract=delegation_contract,
            amount=1000000000000000000
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "unDelegate@0de0b6b3a7640000"
        assert transaction.value == 0
        assert transaction.gas_limit == 12000000

    def test_create_transaction_for_withdrawing(self):
        sender = Address.new_from_bech32("moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6")
        delegation_contract = Address.new_from_bech32("moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g")

        transaction = self.factory.create_transaction_for_withdrawing(
            sender=sender,
            delegation_contract=delegation_contract
        )

        assert transaction.sender == "moa18s6a06ktr2v6fgxv4ffhauxvptssnaqlds45qgsrucemlwc8rawqevhvh6"
        assert transaction.receiver == "moa1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqtlllllszhvd5g"
        assert transaction.data
        assert transaction.data.decode() == "withdraw"
        assert transaction.value == 0
        assert transaction.gas_limit == 12000000



from typing import List, Optional, Protocol, Sequence

from drt_sdk_core.errors import BadUsageError
from drt_sdk_core.interfaces import IAddress, IToken, ITokenTransfer
from drt_sdk_core.transaction import Transaction
from drt_sdk_core.transaction_factories.token_transfers_data_builder import \
    TokenTransfersDataBuilder
from drt_sdk_core.transaction_factories.transaction_builder import \
    TransactionBuilder

ADDITIONAL_GAS_FOR_DCDT_TRANSFER = 100000
ADDITIONAL_GAS_FOR_DCDT_NFT_TRANSFER = 800000


class IConfig(Protocol):
    chain_id: str
    min_gas_limit: int
    gas_limit_per_byte: int
    gas_limit_dcdt_transfer: int
    gas_limit_dcdt_nft_transfer: int
    gas_limit_multi_dcdt_nft_transfer: int


class ITokenComputer(Protocol):
    def is_fungible(self, token: IToken) -> bool:
        ...

    def extract_identifier_from_extended_identifier(self, identifier: str) -> str:
        ...


class TransferTransactionsFactory:
    def __init__(self, config: IConfig, token_computer: ITokenComputer) -> None:
        self.config = config
        self.token_computer = token_computer
        self._data_args_builder = TokenTransfersDataBuilder(token_computer)

    def create_transaction_for_native_token_transfer(self,
                                                     sender: IAddress,
                                                     receiver: IAddress,
                                                     native_amount: int,
                                                     data: Optional[str] = None) -> Transaction:
        transaction_data = data if data else ""
        return TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=receiver,
            data_parts=[transaction_data],
            gas_limit=0,
            add_data_movement_gas=True,
            amount=native_amount
        ).build()

    def create_transaction_for_dcdt_token_transfer(self,
                                                   sender: IAddress,
                                                   receiver: IAddress,
                                                   token_transfers: Sequence[ITokenTransfer]) -> Transaction:
        data_parts: List[str] = []
        extra_gas_for_transfer = 0

        if len(token_transfers) == 0:
            raise BadUsageError("No token transfer has been provided")
        elif len(token_transfers) == 1:
            transfer = token_transfers[0]

            if self.token_computer.is_fungible(transfer.token):
                data_parts = self._data_args_builder.build_args_for_dcdt_transfer(transfer)
                extra_gas_for_transfer = self.config.gas_limit_dcdt_transfer + ADDITIONAL_GAS_FOR_DCDT_TRANSFER
            else:
                data_parts = self._data_args_builder.build_args_for_single_dcdt_nft_transfer(transfer, receiver)
                extra_gas_for_transfer = self.config.gas_limit_dcdt_nft_transfer + ADDITIONAL_GAS_FOR_DCDT_NFT_TRANSFER
                receiver = sender
        else:
            data_parts = self._data_args_builder.build_args_for_multi_dcdt_nft_transfer(receiver, token_transfers)
            extra_gas_for_transfer = self.config.gas_limit_multi_dcdt_nft_transfer * len(token_transfers) + ADDITIONAL_GAS_FOR_DCDT_NFT_TRANSFER
            receiver = sender

        return TransactionBuilder(
            config=self.config,
            sender=sender,
            receiver=receiver,
            data_parts=data_parts,
            gas_limit=extra_gas_for_transfer,
            add_data_movement_gas=True
        ).build()



"""
Runtime type-checking is not a good idea, generally speaking. 
However, some developers may inadvertently disable static type-checking in their IDEs,
and this can lead to hard-to-debug errors.

The functions in this module are meant to be **internally** used by some components,
in order to guard against misuse of the SDK, which may result in loss of funds.
"""

from typing import Any


def assert_is_integer(value: Any, message: str = ""):
    if not isinstance(value, int):
        message = message or f"Expected integer, got {type(value)}"
        raise ValueError(message)
